"""OnboardingTask Django model.

(c) 2020 Network To Code
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.db import models
from django.urls import reverse
from dcim.models import Device
from .choices import OnboardingStatusChoices, OnboardingFailChoices
from .release import NETBOX_RELEASE_CURRENT, NETBOX_RELEASE_29


class OnboardingTask(models.Model):
    """The status of each onboarding Task is tracked in the OnboardingTask table."""

    created_device = models.ForeignKey(to="dcim.Device", on_delete=models.SET_NULL, blank=True, null=True)

    ip_address = models.CharField(max_length=255, help_text="primary ip address for the device", null=True)

    site = models.ForeignKey(to="dcim.Site", on_delete=models.SET_NULL, blank=True, null=True)

    role = models.ForeignKey(to="dcim.DeviceRole", on_delete=models.SET_NULL, blank=True, null=True)

    device_type = models.CharField(
        null=True, max_length=255, help_text="Device Type extracted from the device (optional)"
    )

    platform = models.ForeignKey(to="dcim.Platform", on_delete=models.SET_NULL, blank=True, null=True)

    status = models.CharField(max_length=255, choices=OnboardingStatusChoices, help_text="Overall status of the task")

    failed_reason = models.CharField(
        max_length=255, choices=OnboardingFailChoices, help_text="Raison why the task failed (optional)", null=True
    )

    message = models.CharField(max_length=511, blank=True)

    port = models.PositiveSmallIntegerField(help_text="Port to use to connect to the device", default=22)
    timeout = models.PositiveSmallIntegerField(
        help_text="Timeout period in sec to wait while connecting to the device", default=30
    )

    created_on = models.DateTimeField(auto_now_add=True)

    class Meta:  # noqa: D106 "missing docstring in public nested class"
        ordering = ["created_on"]

    def __str__(self):
        """String representation of an OnboardingTask."""
        return f"{self.site} : {self.ip_address}"

    def get_absolute_url(self):
        """Provide absolute URL to an OnboardingTask."""
        return reverse("plugins:netbox_onboarding:onboardingtask", kwargs={"pk": self.pk})

    if NETBOX_RELEASE_CURRENT >= NETBOX_RELEASE_29:
        from utilities.querysets import RestrictedQuerySet  # pylint: disable=no-name-in-module, import-outside-toplevel

        objects = RestrictedQuerySet.as_manager()


class OnboardingDevice(models.Model):
    """The status of each Onboarded Device is tracked in the OnboardingDevice table."""

    device = models.OneToOneField(to="dcim.Device", on_delete=models.CASCADE)
    enabled = models.BooleanField(default=True, help_text="Whether (re)onboarding of this device is permitted")

    @property
    def last_check_attempt_date(self):
        """Date of last onboarding attempt for a device."""
        try:
            return OnboardingTask.objects.filter(created_device=self.device).latest("created_on").created_on
        except ValueError:
            return "unknown"

    @property
    def last_check_successful_date(self):
        """Date of last successful onboarding for a device."""
        try:
            return (
                OnboardingTask.objects.filter(
                    created_device=self.device, status=OnboardingStatusChoices.STATUS_SUCCEEDED
                )
                .latest("created_on")
                .created_on
            )
        except ValueError:
            return "unknown"

    @property
    def status(self):
        """Last onboarding status."""
        try:
            return OnboardingTask.objects.filter(created_device=self.device).latest("created_on").status
        except ValueError:
            return "unknown"

    @property
    def last_ot(self):
        """Last onboarding task."""
        try:
            return OnboardingTask.objects.filter(created_device=self.device).latest("created_on")
        except ValueError:
            return None


@receiver(post_save, sender=Device)
def init_onboarding_for_new_device(sender, instance, created, **kwargs):  # pylint: disable=unused-argument
    """Register to create a OnboardingDevice object for each new Device Object using Django Signal.

    https://docs.djangoproject.com/en/3.0/ref/signals/#post-save
    """
    if created:
        OnboardingDevice.objects.create(device=instance)



"""Onboard.

(c) 2020 Network To Code
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from django.conf import settings

from .netdev_keeper import NetdevKeeper

PLUGIN_SETTINGS = settings.PLUGINS_CONFIG["netbox_onboarding"]


class OnboardingTaskManager:
    """Onboarding Task Manager."""

    def __init__(self, ot):
        """Inits class."""
        self.ot = ot

    @property
    def napalm_driver(self):
        """Return napalm driver name."""
        if self.ot.platform and self.ot.platform.napalm_driver:
            return self.ot.platform.napalm_driver

        return None

    @property
    def ip_address(self):
        """Return ot's ip address."""
        return self.ot.ip_address

    @property
    def port(self):
        """Return ot's port."""
        return self.ot.port

    @property
    def timeout(self):
        """Return ot's timeout."""
        return self.ot.timeout

    @property
    def site(self):
        """Return ot's site."""
        return self.ot.site

    @property
    def device_type(self):
        """Return ot's device type."""
        return self.ot.device_type

    @property
    def role(self):
        """Return it's device role."""
        return self.ot.role

    @property
    def platform(self):
        """Return ot's device platform."""
        return self.ot.platform


class OnboardingManager:
    """Onboarding Manager."""

    def __init__(self, ot, username, password, secret):
        """Inits class."""
        self.username = username
        self.password = password
        self.secret = secret

        # Create instance of Onboarding Task Manager class:
        otm = OnboardingTaskManager(ot)

        netdev = NetdevKeeper(
            hostname=otm.ip_address,
            port=otm.port,
            timeout=otm.timeout,
            username=self.username,
            password=self.password,
            secret=self.secret,
            napalm_driver=otm.napalm_driver,
        )

        netdev.get_onboarding_facts()
        netdev_dict = netdev.get_netdev_dict()

        onboarding_kwargs = {
            # Kwargs extracted from OnboardingTask:
            "netdev_mgmt_ip_address": otm.ip_address,
            "netdev_nb_site_slug": otm.site.slug if otm.site else None,
            "netdev_nb_device_type_slug": otm.device_type,
            "netdev_nb_role_slug": otm.role.slug if otm.role else PLUGIN_SETTINGS["default_device_role"],
            "netdev_nb_role_color": PLUGIN_SETTINGS["default_device_role_color"],
            "netdev_nb_platform_slug": otm.platform.slug if otm.platform else None,
            # Kwargs discovered on the Onboarded Device:
            "netdev_hostname": netdev_dict["netdev_hostname"],
            "netdev_vendor": netdev_dict["netdev_vendor"],
            "netdev_model": netdev_dict["netdev_model"],
            "netdev_serial_number": netdev_dict["netdev_serial_number"],
            "netdev_mgmt_ifname": netdev_dict["netdev_mgmt_ifname"],
            "netdev_mgmt_pflen": netdev_dict["netdev_mgmt_pflen"],
            "netdev_netmiko_device_type": netdev_dict["netdev_netmiko_device_type"],
            "onboarding_class": netdev_dict["onboarding_class"],
            "driver_addon_result": netdev_dict["driver_addon_result"],
        }

        onboarding_cls = netdev_dict["onboarding_class"]()
        onboarding_cls.run(onboarding_kwargs=onboarding_kwargs)

        self.created_device = onboarding_cls.created_device



"""Unit tests for netbox_onboarding views.

(c) 2020 Network To Code
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from dcim.models import Site
from utilities.testing import ViewTestCases

from netbox_onboarding.models import OnboardingTask
from netbox_onboarding.release import NETBOX_RELEASE_CURRENT, NETBOX_RELEASE_29


if NETBOX_RELEASE_CURRENT >= NETBOX_RELEASE_29:

    class OnboardingTestCase(
        ViewTestCases.GetObjectViewTestCase,
        ViewTestCases.ListObjectsViewTestCase,
        ViewTestCases.CreateObjectViewTestCase,
        ViewTestCases.BulkDeleteObjectsViewTestCase,
        ViewTestCases.BulkImportObjectsViewTestCase,  # pylint: disable=no-member
    ):
        """Test the OnboardingTask views."""

        def _get_base_url(self):
            return "plugins:{}:{}_{{}}".format(self.model._meta.app_label, self.model._meta.model_name)

        model = OnboardingTask

        @classmethod
        def setUpTestData(cls):  # pylint: disable=invalid-name, missing-function-docstring
            """Setup test data."""
            site = Site.objects.create(name="USWEST", slug="uswest")
            OnboardingTask.objects.create(ip_address="10.10.10.10", site=site)
            OnboardingTask.objects.create(ip_address="192.168.1.1", site=site)

            cls.form_data = {
                "site": site.pk,
                "ip_address": "192.0.2.99",
                "port": 22,
                "timeout": 30,
            }

            cls.csv_data = (
                "site,ip_address",
                "uswest,10.10.10.10",
                "uswest,10.10.10.20",
                "uswest,10.10.10.30",
            )



"""User credentials helper module for device onboarding.

(c) 2020 Network To Code
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class Credentials:
    """Class used to hide user's credentials in RQ worker and Django."""

    def __init__(self, username=None, password=None, secret=None):
        """Create a Credentials instance."""
        self.username = username
        self.password = password
        self.secret = secret

    def __repr__(self):
        """Return string representation of a Credentials object."""
        return "*Credentials argument hidden*"



"""Unit tests for netbox_onboarding.netdev_keeper module and its classes.

(c) 2020 Network To Code
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from socket import gaierror
from unittest import mock

from django.test import TestCase
from dcim.models import Site, DeviceRole, Platform

from netbox_onboarding.exceptions import OnboardException
from netbox_onboarding.models import OnboardingTask
from netbox_onboarding.netdev_keeper import NetdevKeeper


class NetdevKeeperTestCase(TestCase):
    """Test the NetdevKeeper Class."""

    def setUp(self):
        """Create a superuser and token for API calls."""
        self.site1 = Site.objects.create(name="USWEST", slug="uswest")
        self.device_role1 = DeviceRole.objects.create(name="Firewall", slug="firewall")

        self.platform1 = Platform.objects.create(name="JunOS", slug="junos", napalm_driver="junos")
        # self.platform2 = Platform.objects.create(name="Cisco NX-OS", slug="cisco-nx-os")

        self.onboarding_task4 = OnboardingTask.objects.create(
            ip_address="ntc123.local", site=self.site1, role=self.device_role1, platform=self.platform1
        )

        self.onboarding_task5 = OnboardingTask.objects.create(
            ip_address="bad.local", site=self.site1, role=self.device_role1, platform=self.platform1
        )

        self.onboarding_task7 = OnboardingTask.objects.create(
            ip_address="192.0.2.1/32", site=self.site1, role=self.device_role1, platform=self.platform1
        )

    @mock.patch("netbox_onboarding.netdev_keeper.socket.gethostbyname")
    def test_check_ip(self, mock_get_hostbyname):
        """Check DNS to IP address."""
        # Look up response value
        mock_get_hostbyname.return_value = "192.0.2.1"

        # Create a Device Keeper object of the device
        ndk4 = NetdevKeeper(hostname=self.onboarding_task4.ip_address)

        # Check that the IP address is returned
        self.assertTrue(ndk4.check_ip())

        # Run the check to change the IP address
        self.assertEqual(ndk4.hostname, "192.0.2.1")

    @mock.patch("netbox_onboarding.netdev_keeper.socket.gethostbyname")
    def test_failed_check_ip(self, mock_get_hostbyname):
        """Check DNS to IP address failing."""
        # Look up a failed response
        mock_get_hostbyname.side_effect = gaierror(8)
        ndk5 = NetdevKeeper(hostname=self.onboarding_task5.ip_address)
        ndk7 = NetdevKeeper(hostname=self.onboarding_task7.ip_address)

        # Check for bad.local raising an exception
        with self.assertRaises(OnboardException) as exc_info:
            ndk5.check_ip()
            self.assertEqual(exc_info.exception.message, "ERROR failed to complete DNS lookup: bad.local")
            self.assertEqual(exc_info.exception.reason, "fail-dns")

        # Check for exception with prefix address entered
        with self.assertRaises(OnboardException) as exc_info:
            ndk7.check_ip()
            self.assertEqual(exc_info.exception.reason, "fail-prefix")
            self.assertEqual(exc_info.exception.message, "ERROR appears a prefix was entered: 192.0.2.1/32")
